import codecs
import logging
import json
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

from . import exceptions, config
from .storages import StorageMapper, EnvFile, partition_path

logger = logging.getLogger(__name__)
__escape_decoder = codecs.getdecoder('unicode_escape')


def parse_dotenv(data):
    for line in data.splitlines():
        line = line.strip()
        if line and not line.startswith('#') and '=' in line:
            k, _, v = line.partition('=')

            # Remove any leading and trailing spaces in key, value
            k, v = k.strip(), v.strip().encode('unicode-escape').decode('ascii')

            if v and v[0] == v[-1] in ['"', "'"]:
                v = __escape_decoder(v[1:-1])[0]

            yield k, v


def phusion_dump(environment, path):
    path = Path(path)
    path.mkdir(parents=True, exist_ok=True)
    for k, v in environment.items():
        with open(path.joinpath(k), 'w') as f:
            f.write(v + '\n')


def raise_out_of_sync(local_file, remote_file):
    raise exceptions.LocalCopyOutdated(
        'Upload failed, remote file probably changed since last download and md5 hashes in cache conflict.\n'
        'If you want to upload anyway, use the -f flag. \n'
        'Offending file:\n\n    %s -> %s ',
        local_file,
        remote_file
    )


class S3Conf:
    def __init__(self, settings=None):
        self.settings = settings or config.Settings()
        self._storages = None

    @property
    def storages(self):
        if not self._storages:
            self._storages = StorageMapper(self.settings)
        return self._storages

    def verify_cache(self):
        cache_dir = Path(self.settings.cache_dir)
        cache_dir.mkdir(parents=True, exist_ok=True)
        default_config_file = config.ConfigFileResolver(self.settings.default_config_file, section='DEFAULT')
        default_config_file.save()
        gitignore_file_path = cache_dir.joinpath('.gitignore')
        open(gitignore_file_path, 'w').write('*\n')

    def check_remote_changes(self):
        local_hashes = json.load(open(self.settings.hash_file)) if Path(self.settings.hash_file).exists() else {}
        for local_path, remote_path in self.settings.file_mappings.items():
            remote_hashes = self.storages.list(remote_path)
            for _, local_file, remote_file in self.storages.expand_path(local_path, remote_path):
                local_hash = local_hashes.get(str(local_file))
                remote_hash = remote_hashes.get(str(remote_file))
                if local_hash:
                    if local_hash != remote_hash:
                        raise_out_of_sync(local_file, remote_file)
                else:
                    logger.warning('New mapped file detected: %s', local_file)

    def push(self, force=False):
        if not force:
            self.check_remote_changes()
        hashes = {}

        with ThreadPoolExecutor() as executor:
            futures = []
            for local_path, remote_path in self.settings.file_mappings.items():
                futures.append(executor.submit(self.storages.copy, local_path, remote_path))
            for future in as_completed(futures):
                copy_hashes = future.result()
                hashes.update({str(local_file): local_hash for local_hash, local_file, _ in copy_hashes})

        self.verify_cache()
        json.dump(hashes, open(self.settings.hash_file, 'w'), indent=4)
        return hashes

    def pull(self):
        hashes = {}

        with ThreadPoolExecutor() as executor:
            futures = []
            for local_path, remote_path in self.settings.file_mappings.items():
                futures.append(executor.submit(self.storages.copy, remote_path, local_path))
            for future in as_completed(futures):
                copy_hashes = future.result()
                hashes.update({str(local_file): remote_hash for remote_hash, _, local_file in copy_hashes})

        self.verify_cache()
        json.dump(hashes, open(self.settings.hash_file, 'w'), indent=4)
        return hashes

    def get_envfile(self, mode='r'):
        logger.info('Loading configs from %s', self.settings.environment_file_path)
        remote_storage = self.storages.storage(self.settings.environment_file_path)
        _, _, path = partition_path(self.settings.environment_file_path)
        return EnvFile.from_file(remote_storage.open(path, mode=mode))

    def edit_envfile(self):
        with self.get_envfile(mode='r+') as envfile:
            envfile.edit()

    def create_envfile(self):
        logger.info('Trying to create envifile %s', self.settings.environment_file_path)
        remote_storage = self.storages.storage(self.settings.environment_file_path)
        remote_storage.create_bucket()
        _, _, path = partition_path(self.settings.environment_file_path)
        envfile_exist = bool(list(remote_storage.list(path)))
        if envfile_exist:
            logger.warning('%s already exist', self.settings.environment_file_path)
        else:
            with self.get_envfile(mode='w') as _:
                pass



import os
import codecs
import logging
import difflib
from concurrent.futures import ThreadPoolExecutor
from shutil import copyfileobj
from pathlib import Path
from functools import lru_cache

import editor

from .storage.storages import S3Storage, GCStorage, LocalStorage, s3etag
from .storage.files import File
from .storage.exceptions import FileDoesNotExist

logger = logging.getLogger(__name__)
__escape_decoder = codecs.getdecoder('unicode_escape')


def parse_env_var(value):
    """
    Split a env var text like

    ENV_VAR_NAME=env_var_value

    into a tuple ('ENV_VAR_NAME', 'env_var_value')
    """
    k, _, v = value.partition('=')

    # Remove any leading and trailing spaces in key, value
    k, v = k.strip(), v.strip().encode('unicode-escape').decode('ascii')

    if v and v[0] == v[-1] in ['"', "'"]:
        v = __escape_decoder(v[1:-1])[0]
    return k, v


def strip_prefix(text, prefix):
    return text[len(prefix):] if text.startswith(prefix) else text


def partition_path(path):
    protocol, _, path = str(path).partition(':')
    if not path:
        path = protocol
        protocol = 'file'
        bucket = None
    else:
        bucket, _, path = path.lstrip('/').partition('/')
    return protocol, bucket, path


def build_path(protocol, bucket, path):
    if protocol == 'file':
        return path
    else:
        return f'{protocol}://{bucket}/{path}'


def get_s3_storage(settings, file_class, bucket):
    return S3Storage(
        aws_access_key_id=settings.get('S3CONF_ACCESS_KEY_ID') or settings.get('AWS_ACCESS_KEY_ID'),
        aws_secret_access_key=settings.get('S3CONF_SECRET_ACCESS_KEY') or settings.get('AWS_SECRET_ACCESS_KEY'),
        aws_session_token=settings.get('S3CONF_SESSION_TOKEN') or settings.get('AWS_SESSION_TOKEN'),
        region_name=settings.get('S3CONF_S3_REGION_NAME') or settings.get('AWS_S3_REGION_NAME'),
        use_ssl=settings.get('S3CONF_S3_USE_SSL') or settings.get('AWS_S3_USE_SSL', True),
        endpoint_url=settings.get('S3CONF_S3_ENDPOINT_URL') or settings.get('AWS_S3_ENDPOINT_URL'),
        file_class=file_class,
        bucket=bucket,
    )


def get_gs_storage(settings, file_class, bucket):
    return GCStorage(
        credential_file=settings.get('S3CONF_APPLICATION_CREDENTIALS') or settings.get(
            'GOOGLE_APPLICATION_CREDENTIALS'),
        file_class=file_class,
        bucket=bucket,
    )


def list_all_files(path):
    if Path(path).is_dir():
        mapping = iter(str(Path(root).joinpath(name).resolve()) for root, _, names in os.walk(path) for name in names)
    else:
        mapping = [str(path)]
    return mapping


@lru_cache()
def _make_storage(mapper, protocol, bucket):
    try:
        storage = {
            's3': get_s3_storage(mapper.settings, BaseFile, bucket),
            'gs': get_gs_storage(mapper.settings, BaseFile, bucket),
        }[protocol]
    except KeyError:
        storage = LocalStorage(file_class=BaseFile, root=mapper.settings.root_folder, hash_method=mapper.local_hash)
    logger.debug('Using %s storage', storage)
    return storage


class StorageMapper:
    def __init__(self, settings):
        self.settings = settings
        self.local_hash = self.storage(self.settings.environment_file_path).hash_method

    def storage(self, path=''):
        protocol, bucket, _ = partition_path(path)
        logger.debug('Getting storage for %s %s', protocol, bucket)
        return _make_storage(self, protocol, bucket)

    def expand_path(self, source_path, target_path):
        source = self.storage(source_path)
        source_protocol, source_bucket, source_path = partition_path(source_path)
        target_protocol, target_bucket, target_path = partition_path(target_path)
        files = {file_path: source_hash for source_hash, file_path in source.list(source_path)}
        is_dir = False if len(files) == 1 and source_path in files else True
        if not is_dir:
            yield files[source_path], \
                  build_path(source_protocol, source_bucket, source_path), \
                  build_path(target_protocol, target_bucket, target_path)
        else:
            for source_file_path, source_hash in files.items():
                target_file_path = os.path.join(target_path, os.path.relpath(source_file_path, source_path))
                yield source_hash, \
                      build_path(source_protocol, source_bucket, source_file_path), \
                      build_path(target_protocol, target_bucket, target_file_path)

    def list(self, path):
        target = self.storage(path)
        protocol, bucket, sripped_path = partition_path(path)
        return {build_path(protocol, bucket, file_path): etag for etag, file_path in target.list(sripped_path)}

    def prepare_copy_list(self, source_path, target_path, force=False):
        to_copy = []
        target_hashes = {}
        final_state = []

        # if we are not forcing the copy, we check the hashes to avoid unnecessary copies
        if not force:
            target_hashes = self.list(target_path)

        for source_hash, source_file, target_file in self.expand_path(source_path, target_path):
            if target_file not in target_hashes or target_hashes[target_file] != source_hash:
                to_copy.append((source_hash, source_file, target_file))
            else:
                logger.debug('Skipping (no changes): %s -> %s', source_file, target_file)
            final_state.append((source_hash, source_file, target_file))

        return final_state, to_copy

    def _copy(self, source, target, source_file, target_file):
        _, _, source_file = partition_path(source_file)
        _, _, target_file = partition_path(target_file)
        with source.open(source_file) as source_stream, target.open(target_file, 'wb') as target_stream:
            copyfileobj(source_stream, target_stream)

    def copy(self, source_path, target_path, force=False):
        final_state, copy_list = self.prepare_copy_list(source_path, target_path, force)
        source = self.storage(source_path)
        target = self.storage(target_path)

        with ThreadPoolExecutor() as executor:
            for _, source_file, target_file in copy_list:
                executor.submit(self._copy, source, target, source_file, target_file)

        return final_state


class BaseFile(File):
    def exists(self):
        if list(self.storage.list(self.name)):
            return True
        return False

    def md5(self, raise_if_not_exists=True):
        try:
            md5hash, _ = next(self.storage.list(self.name))
        except StopIteration:
            if raise_if_not_exists:
                raise FileDoesNotExist(self.name)
            else:
                md5hash = None
        return md5hash

    def diff(self, file_stream, fromfile='remote', tofile='local', **kwargs):
        file_stream.seek(0)
        self.seek(0)
        result = difflib.unified_diff(
            self.readlines(),
            file_stream.readlines(),
            fromfile=fromfile,
            tofile=tofile,
            **kwargs
        )
        return result

    def edit(self):
        self.seek(0)
        original_data = self.read()
        edited_data = editor.edit(contents=original_data)
        self.seek(0)
        self.truncate()
        self.write(edited_data.decode())


class EnvFile(BaseFile):
    @classmethod
    def from_file(cls, obj):
        obj.__class__ = cls
        return obj

    def as_dict(self):
        self.seek(0)
        lines = self.read().splitlines()
        lines = [line for line in lines if line and not line.startswith('#') and '=' in line]
        env_dict = dict(parse_env_var(line) for line in lines)
        return env_dict

    def set(self, value):
        new_key, new_value = parse_env_var(value)
        new_lines = []
        value_set = False
        self.seek(0)
        for line in self.read().splitlines():
            key, value = parse_env_var(line)
            if key == new_key:
                new_lines.append('{}={}'.format(new_key, new_value))
                value_set = True
            else:
                new_lines.append(line)
        if not value_set:
            new_lines.append('{}={}'.format(new_key, new_value))
        self.seek(0)
        self.truncate()
        self.write('\n'.join(new_lines))
        self.file.flush()

    def unset(self, unset_key):
        new_lines = []
        unset_done = False
        try:
            self.seek(0)
            for line in self.read().splitlines():
                key, value = parse_env_var(line)
                if key == unset_key:
                    unset_done = True
                    continue
                else:
                    new_lines.append(line)
        except FileNotFoundError:
            logger.warning('File does not exist')

        if unset_done:
            self.seek(0)
            self.truncate()
            self.write('\n'.join(new_lines))
            self.file.flush()
        else:
            logger.info('Key %s not found in environemnt file, doing nothing...', unset_key)

    def from_dict(self, env_vars):
        file_data = ''
        for var_name, var_value in env_vars.items():
            file_data += '{}={}\n'.format(var_name, var_value)
        self.truncate()
        self.write(file_data)

    def update(self, another_file):
        env_vars = self.as_dict()
        env_vars.update(another_file.as_dict())
        self.from_dict(env_vars)



import os
import logging
from configobj import ConfigObj
from pathlib import Path

from . import exceptions


logger = logging.getLogger(__name__)


CONFIG_NAME = 's3conf'


def _lookup_root_folder(current_path='.'):
    current_path = Path(current_path).resolve()
    # recursion stops
    if not current_path or current_path == Path(current_path.anchor):
        root_folder = Path('.')
        logger.debug('Root folder detected: %s', root_folder)
        return root_folder
    path_items = {entry.name: entry for entry in os.scandir(current_path)}
    config_file_name = f'{CONFIG_NAME}.ini'
    if config_file_name in path_items:
        entry = path_items[config_file_name]
        if entry.is_file():
            logger.debug('Root folder detected: %s', current_path)
            return str(current_path)
    return _lookup_root_folder(current_path.parent)


class EnvironmentResolver:
    def get(self, item, default=None):
        return os.environ.get(item, default)

    def __str__(self):
        return f'environemnt'


class ConfigFileResolver:
    def __init__(self, config_file, section=None):
        self.config_file = config_file
        self.section = section or 'DEFAULT'
        self._config = None

    def __str__(self):
        return f'{self.config_file}:{self.section}'

    @property
    def config(self):
        if not self._config:
            self._config = ConfigObj(str(self.config_file))
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    def get(self, item, default=None, section=None):
        try:
            return self.config[section or self.section][item]
        except KeyError:
            return default

    def set(self, item, value, section=None):
        self.config.setdefault(section or self.section, {})[item] = value

    def save(self):
        self.config.write()

    def sections(self):
        return list(self.config)


class Settings:
    def __init__(self, section=None, config_file=None):
        if config_file:
            config_file = Path(config_file).resolve()
            root_folder = Path(config_file).parent
        else:
            root_folder = Path(_lookup_root_folder()).resolve()
            config_file = root_folder.joinpath(f'{CONFIG_NAME}.ini')
        self.root_folder = str(root_folder)
        self.config_file = str(config_file)
        cache_dir = root_folder.joinpath(f'.{CONFIG_NAME}')
        self.cache_dir = str(cache_dir)
        self.hash_file = str(cache_dir.joinpath('remote_hashes'))
        self.default_config_file = str(cache_dir.joinpath('default.ini'))
        self.section = section
        logger.debug('Settings paths:\n%s\n%s\n%s\n%s',
                     self.root_folder,
                     self.config_file,
                     self.cache_dir,
                     self.default_config_file)

        if self.section:
            self.resolvers = [
                ConfigFileResolver(self.config_file, self.section),
                EnvironmentResolver(),
                ConfigFileResolver(self.default_config_file),
            ]
        else:
            self.resolvers = [
                EnvironmentResolver(),
                ConfigFileResolver(self.config_file, self.section),
                ConfigFileResolver(self.default_config_file),
            ]

        self._environment_file_path = None
        self._file_mappings = None

    @property
    def environment_file_path(self):
        # resolving environment file path
        if not self._environment_file_path:
            self._environment_file_path = self.get('S3CONF')
            if not self._environment_file_path:
                logger.error('Environemnt file name is not defined or is empty.')
                raise exceptions.EnvfilePathNotDefinedError()
        return self._environment_file_path

    @property
    def file_mappings(self):
        if self._file_mappings is None:
            self._file_mappings = {}
            files_list = self.get('S3CONF_MAP')
            files_pairs = files_list.split(';') if files_list else []
            for file_map in files_pairs:
                remote_path, _, local_path = file_map.rpartition(':')
                if remote_path and local_path:
                    self.add_mapping(remote_path, local_path)
        return self._file_mappings

    def add_mapping(self, remote_path, local_path):
        local_path = Path(local_path)
        local_path = Path(self.root_folder).joinpath(local_path.relative_to(local_path.root))
        self.file_mappings[str(local_path)] = remote_path

    def rm_mapping(self, local_path):
        local_path = Path(local_path)
        local_path = self.root_folder.joinpath(local_path.relative_to(local_path.root))
        del self.file_mappings[str(local_path)]

    def serialize_mappings(self):
        file_mappings = {}
        for local_path, remote_path in self.file_mappings.items():
            relative_local_path = Path(local_path).relative_to(self.root_folder)
            file_mappings[str(relative_local_path)] = remote_path
        return ';'.join(f'{remote_path}:{local_path}' for local_path, remote_path in file_mappings.items())

    def __getitem__(self, item):
        for resolver in self.resolvers:
            value = resolver.get(item)
            if value:
                break
        else:
            raise KeyError()
        logger.debug('Entry %s has value %s found in %s', item, value, str(resolver))
        return value

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            logger.debug('Key %s not found, returning default %s', item, default)
            return default



import functools
import logging
from tempfile import NamedTemporaryFile

from . import exceptions


logger = logging.getLogger(__name__)


class File:
    def __init__(self, name, storage, mode='r', encoding=None):
        if 'x' in mode:
            raise ValueError('Exclusive creation not supported yet.')
        self.name = name
        self.storage = storage
        self.mode = mode
        self.encoding = encoding
        self._buffer = None
        self._file = None

    def _sync_with_storage(self):
        if 'w' not in self.mode:
            try:
                self._buffer.seek(0)
                self._buffer.truncate()
                self.storage.read_into_stream(self.name, self._buffer)
            except exceptions.FileDoesNotExist:
                self.file.close()
                raise

    # IOBase
    def flush(self) -> None:
        if self.file.writable():
            self.file.flush()
            self._buffer.seek(0)
            logger.debug('Writing buffer to %s', self.name)
            self.storage.write(self._buffer, self.name)

    # IOBase
    def close(self) -> None:
        if not self.closed:
            self.flush()
            self.file.close()
            self._buffer.close()

    @property
    def file(self):
        if self._file is None:
            self._buffer = NamedTemporaryFile()
            self._file = open(self._buffer.name, mode=self.mode, encoding=self.encoding)
            self._sync_with_storage()
        return self._file

    # Inspired by
    # https://github.com/python/cpython/blob/29500737d45cbca9604d9ce845fb2acc3f531401/Lib/tempfile.py#L461
    def __getattr__(self, name):
        # Attribute lookups are delegated to the underlying file
        # and cached for non-numeric results
        # (i.e. methods are cached, closed and friends are not)
        a = getattr(self.file, name)
        if hasattr(a, '__call__'):
            func = a
            @functools.wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            a = func_wrapper
        if not isinstance(a, int):
            setattr(self, name, a)
        return a

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    # IOBase
    def fileno(self) -> int:
        raise OSError()
